"""Databricks SDK integration.

Databricks SDK credentials implementations which support interacting with Posit OAuth integrations on Connect.

Notes
-----
These APIs are provided as a convenience and are subject to breaking changes:
https://github.com/databricks/databricks-sdk-py#interface-stability
"""

import abc
from typing import Callable, Dict, Optional

import requests

from ..client import Client
from ..oauth import Credentials
from .external import is_local

POSIT_OAUTH_INTEGRATION_AUTH_TYPE = "posit-oauth-integration"
POSIT_LOCAL_CLIENT_CREDENTIALS_AUTH_TYPE = "posit-local-client-credentials"

# The Databricks SDK CredentialsProvider == Databricks SQL HeaderFactory
CredentialsProvider = Callable[[], Dict[str, str]]


class CredentialsStrategy(abc.ABC):
    """Maintain compatibility with the Databricks SQL/SDK client libraries.

    See Also
    --------
    * https://github.com/databricks/databricks-sql-python/blob/v3.3.0/src/databricks/sql/auth/authenticators.py#L19-L33
    * https://github.com/databricks/databricks-sdk-py/blob/v0.29.0/databricks/sdk/credentials_provider.py#L44-L54
    """

    @abc.abstractmethod
    def auth_type(self) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def __call__(self, *args, **kwargs) -> CredentialsProvider:
        raise NotImplementedError


def _new_bearer_authorization_header(credentials: Credentials) -> Dict[str, str]:
    """Helper to transform an Credentials object into the Bearer auth header consumed by databricks.

    Raises
    ------
        ValueError: If provided Credentials object does not contain an access token

    Returns
    -------
        Dict[str, str]
    """
    access_token = credentials.get("access_token")
    if access_token is None:
        raise ValueError("Missing value for field 'access_token' in credentials.")
    return {"Authorization": f"Bearer {access_token}"}


def _get_auth_type(local_auth_type: str) -> str:
    """Returns the auth type currently in use.

    The databricks-sdk client uses the configured auth_type to create
    a user-agent string which is used for attribution. We should only
    overwrite the auth_type if we are using the PositCredentialsStrategy (non-local),
    otherwise, we should return the auth_type of the configured local_strategy instead
    to avoid breaking someone elses attribution.

    NOTE: The databricks-sql client does not use auth_type to set the user-agent.
    https://github.com/databricks/databricks-sql-python/blob/v3.3.0/src/databricks/sql/client.py#L214-L219

    See Also
    --------
    * https://github.com/databricks/databricks-sdk-py/blob/v0.29.0/databricks/sdk/config.py#L261-L269

    Returns
    -------
        str
    """
    if is_local():
        return local_auth_type

    return POSIT_OAUTH_INTEGRATION_AUTH_TYPE


class PositLocalContentCredentialsProvider:
    """`CredentialsProvider` implementation which provides a fallback for local development using a client credentials flow.

    There is an open issue against the Databricks CLI which prevents it from returning service principal access tokens.
    https://github.com/databricks/cli/issues/1939

    Until the CLI issue is resolved, this CredentialsProvider implements the approach described in the Databricks documentation
    for manually generating a workspace-level access token using OAuth M2M authentication. Once it has acquired an access token,
    it returns it as a Bearer authorization header like other `CredentialsProvider` implementations.

    See Also
    --------
    * https://docs.databricks.com/en/dev-tools/auth/oauth-m2m.html#manually-generate-a-workspace-level-access-token
    """

    def __init__(self, token_endpoint_url: str, client_id: str, client_secret: str):
        self._token_endpoint_url = token_endpoint_url
        self._client_id = client_id
        self._client_secret = client_secret

    def __call__(self) -> Dict[str, str]:
        response = requests.post(
            self._token_endpoint_url,
            auth=(self._client_id, self._client_secret),
            data={
                "grant_type": "client_credentials",
                "scope": "all-apis",
            },
        )
        response.raise_for_status()

        credentials = Credentials(**response.json())
        return _new_bearer_authorization_header(credentials)


class PositContentCredentialsProvider:
    """`CredentialsProvider` implementation which initiates a credential exchange using a content-session-token.

    The content-session-token is provided by Connect through the environment variable `CONNECT_CONTENT_SESSION_TOKEN`.

    See Also
    --------
    * https://github.com/posit-dev/posit-sdk-py/blob/main/src/posit/connect/oauth/oauth.py

    """

    def __init__(self, client: Client):
        self._client = client

    def __call__(self) -> Dict[str, str]:
        credentials = self._client.oauth.get_content_credentials()
        return _new_bearer_authorization_header(credentials)


class PositCredentialsProvider:
    """`CredentialsProvider` implementation which initiates a credential exchange using a user-session-token.

    The user-session-token is provided by Connect through the HTTP session header
    `Posit-Connect-User-Session-Token`.

    See Also
    --------
    * https://github.com/posit-dev/posit-sdk-py/blob/main/src/posit/connect/oauth/oauth.py

    """

    def __init__(self, client: Client, user_session_token: str):
        self._client = client
        self._user_session_token = user_session_token

    def __call__(self) -> Dict[str, str]:
        credentials = self._client.oauth.get_credentials(self._user_session_token)
        return _new_bearer_authorization_header(credentials)


class PositLocalContentCredentialsStrategy(CredentialsStrategy):
    """`CredentialsStrategy` implementation which supports local development using OAuth M2M authentication against Databricks.

    There is an open issue against the Databricks CLI which prevents it from returning service principal access tokens.
    https://github.com/databricks/cli/issues/1939

    Until the CLI issue is resolved, this CredentialsStrategy provides a drop-in replacement as a local_strategy that can be used
    to develop applications which target Service Account OAuth integrations on Connect.

    Examples
    --------
    In the example below, the `PositContentCredentialsStrategy` can be initialized anywhere that
    the Python process can read environment variables.

    CLIENT_ID and CLIENT_SECRET are credentials associated with the Databricks service principal.

    ```python
    from posit.connect.external.databricks import (
        PositContentCredentialsStrategy,
        PositLocalContentCredentialsStrategy,
    )

    import pandas as pd
    from databricks import sql
    from databricks.sdk.core import ApiClient, Config
    from databricks.sdk.service.iam import CurrentUserAPI

    DATABRICKS_HOST = "<REDACTED>"
    DATABRICKS_HOST_URL = f"https://{DATABRICKS_HOST}"
    SQL_HTTP_PATH = "<REDACTED>"
    TOKEN_ENDPOINT_URL = f"https://{DATABRICKS_HOST}/oidc/v1/token"

    CLIENT_ID = "<REDACTED>"
    CLIENT_SECRET = "<REDACTED>"

    # Rather than relying on the Databricks CLI as a local strategy, we use
    # PositLocalContentCredentialsStrategy as a drop-in replacement.
    # Can be replaced with the Databricks CLI implementation when
    # https://github.com/databricks/cli/issues/1939 is resolved.
    local_strategy = PositLocalContentCredentialsStrategy(
        token_endpoint_url=TOKEN_ENDPOINT_URL,
        client_id=CLIENT_ID,
        client_secret=CLIENT_SECRET,
    )

    posit_strategy = PositContentCredentialsStrategy(local_strategy=local_strategy)

    cfg = Config(host=DATABRICKS_HOST_URL, credentials_strategy=posit_strategy)

    databricks_user_info = CurrentUserAPI(ApiClient(cfg)).me()
    print(f"Hello, {databricks_user_info.display_name}!")

    query = "SELECT * FROM samples.nyctaxi.trips LIMIT 10;"
    with sql.connect(
        server_hostname=DATABRICKS_HOST,
        http_path=SQL_HTTP_PATH,
        credentials_provider=posit_strategy.sql_credentials_provider(cfg),
    ) as connection:
        with connection.cursor() as cursor:
            cursor.execute(query)
            rows = cursor.fetchall()
            print(pd.DataFrame([row.asDict() for row in rows]))
    ```

    See Also
    --------
    * https://docs.databricks.com/en/dev-tools/auth/oauth-m2m.html#manually-generate-a-workspace-level-access-token
    """

    def __init__(self, token_endpoint_url: str, client_id: str, client_secret: str):
        self._token_endpoint_url = token_endpoint_url
        self._client_id = client_id
        self._client_secret = client_secret

    def sql_credentials_provider(self, *args, **kwargs):
        return lambda: self.__call__(*args, **kwargs)

    def auth_type(self) -> str:
        return POSIT_LOCAL_CLIENT_CREDENTIALS_AUTH_TYPE

    def __call__(self, *args, **kwargs) -> CredentialsProvider:  # noqa: ARG002
        return PositLocalContentCredentialsProvider(
            self._token_endpoint_url,
            self._client_id,
            self._client_secret,
        )


class PositContentCredentialsStrategy(CredentialsStrategy):
    """`CredentialsStrategy` implementation which supports interacting with Service Account OAuth integrations on Connect.

    This strategy callable class returns a `PositContentCredentialsProvider` when hosted on Connect, and
    its `local_strategy` strategy otherwise.

    Examples
    --------
    NOTE: in the example below, the `PositContentCredentialsStrategy` can be initialized anywhere that
    the Python process can read environment variables.

    ```python
    from posit.connect.external.databricks import PositContentCredentialsStrategy

    import pandas as pd
    from databricks import sql
    from databricks.sdk.core import ApiClient, Config, databricks_cli
    from databricks.sdk.service.iam import CurrentUserAPI

    DATABRICKS_HOST = "<REDACTED>"
    DATABRICKS_HOST_URL = f"https://{DATABRICKS_HOST}"
    SQL_HTTP_PATH = "<REDACTED>"

    # NOTE: currently the databricks_cli local strategy only supports auth code OAuth flows.
    # https://github.com/databricks/cli/issues/1939
    #
    # This means that the databricks_cli supports local development using the developer's
    # databricks credentials, but not the credentials for a service principal.
    # To fallback to service principal credentials in local development, use
    # `PositLocalContentCredentialsStrategy` as a drop-in replacement.
    posit_strategy = PositContentCredentialsStrategy(local_strategy=databricks_cli)

    cfg = Config(host=DATABRICKS_HOST_URL, credentials_strategy=posit_strategy)

    databricks_user_info = CurrentUserAPI(ApiClient(cfg)).me()
    print(f"Hello, {databricks_user_info.display_name}!")

    query = "SELECT * FROM samples.nyctaxi.trips LIMIT 10;"
    with sql.connect(
        server_hostname=DATABRICKS_HOST,
        http_path=SQL_HTTP_PATH,
        credentials_provider=posit_strategy.sql_credentials_provider(cfg),
    ) as connection:
        with connection.cursor() as cursor:
            cursor.execute(query)
            rows = cursor.fetchall()
            print(pd.DataFrame([row.asDict() for row in rows]))
    ```
    """

    def __init__(
        self,
        local_strategy: CredentialsStrategy,
        client: Optional[Client] = None,
    ):
        self._local_strategy = local_strategy
        self._client = client

    def sql_credentials_provider(self, *args, **kwargs):
        """The sql connector attempts to call the credentials provider w/o any args.

        The SQL client's `ExternalAuthProvider` is not compatible w/ the SDK's implementation of
        `CredentialsProvider`, so create a no-arg lambda that wraps the args defined by the real caller.
        This way we can pass in a databricks `Config` object required by most of the SDK's `CredentialsProvider`
        implementations from where `sql.connect` is called.

        https://github.com/databricks/databricks-sql-python/issues/148#issuecomment-2271561365
        """
        return lambda: self.__call__(*args, **kwargs)

    def auth_type(self) -> str:
        return _get_auth_type(self._local_strategy.auth_type())

    def __call__(self, *args, **kwargs) -> CredentialsProvider:
        # If the content is not running on Connect then fall back to local_strategy
        if is_local():
            return self._local_strategy(*args, **kwargs)

        if self._client is None:
            self._client = Client()

        return PositContentCredentialsProvider(self._client)


class PositCredentialsStrategy(CredentialsStrategy):
    """`CredentialsStrategy` implementation which supports interacting with Viewer OAuth integrations on Connect.

    This strategy callable class returns a `PositCredentialsProvider` when hosted on Connect, and
    its `local_strategy` strategy otherwise.

    Examples
    --------
    NOTE: In the example below, the PositCredentialsProvider *must* be initialized within the context of the
    shiny `server` function, which provides access to the HTTP session headers.

    ```python
    import os

    import pandas as pd
    from databricks import sql
    from databricks.sdk.core import ApiClient, Config, databricks_cli
    from databricks.sdk.service.iam import CurrentUserAPI
    from posit.connect.external.databricks import PositCredentialsStrategy
    from shiny import App, Inputs, Outputs, Session, render, ui

    DATABRICKS_HOST = "<REDACTED>"
    DATABRICKS_HOST_URL = f"https://{DATABRICKS_HOST}"
    SQL_HTTP_PATH = "<REDACTED>"

    app_ui = ui.page_fluid(ui.output_text("text"), ui.output_data_frame("result"))


    def server(i: Inputs, o: Outputs, session: Session):
        # HTTP session headers are available in this context.
        session_token = session.http_conn.headers.get("Posit-Connect-User-Session-Token")
        posit_strategy = PositCredentialsStrategy(
            local_strategy=databricks_cli, user_session_token=session_token
        )
        cfg = Config(host=DATABRICKS_HOST_URL, credentials_strategy=posit_strategy)

        @render.data_frame
        def result():
            query = "SELECT * FROM samples.nyctaxi.trips LIMIT 10;"

            with sql.connect(
                server_hostname=DATABRICKS_HOST,
                http_path=SQL_HTTP_PATH,
                credentials_provider=posit_strategy.sql_credentials_provider(cfg),
            ) as connection:
                with connection.cursor() as cursor:
                    cursor.execute(query)
                    rows = cursor.fetchall()
                    df = pd.DataFrame(rows, columns=[col[0] for col in cursor.description])
                    return df

        @render.text
        def text():
            databricks_user_info = CurrentUserAPI(ApiClient(cfg)).me()
            return f"Hello, {databricks_user_info.display_name}!"


    app = App(app_ui, server)
    ```
    """

    def __init__(
        self,
        local_strategy: CredentialsStrategy,
        client: Optional[Client] = None,
        user_session_token: Optional[str] = None,
    ):
        self._local_strategy = local_strategy
        self._client = client
        self._user_session_token = user_session_token

    def sql_credentials_provider(self, *args, **kwargs):
        """The sql connector attempts to call the credentials provider w/o any args.

        The SQL client's `ExternalAuthProvider` is not compatible w/ the SDK's implementation of
        `CredentialsProvider`, so create a no-arg lambda that wraps the args defined by the real caller.
        This way we can pass in a databricks `Config` object required by most of the SDK's `CredentialsProvider`
        implementations from where `sql.connect` is called.

        See Also
        --------
        * https://github.com/databricks/databricks-sql-python/issues/148#issuecomment-2271561365
        """
        return lambda: self.__call__(*args, **kwargs)

    def auth_type(self) -> str:
        return _get_auth_type(self._local_strategy.auth_type())

    def __call__(self, *args, **kwargs) -> CredentialsProvider:
        # If the content is not running on Connect then fall back to local_strategy
        if is_local():
            return self._local_strategy(*args, **kwargs)

        # If the user-session-token wasn't provided and we're running on Connect then we raise an exception.
        # user_session_token is required to impersonate the viewer.
        if self._user_session_token is None:
            raise ValueError("The user-session-token is required for viewer authentication.")

        if self._client is None:
            self._client = Client()

        return PositCredentialsProvider(self._client, self._user_session_token)



from email.contentmanager import ContentManager
from unittest.mock import MagicMock, Mock

import pytest
import responses

from posit.connect.client import Client
from posit.connect.context import Context, requires


class TestRequires:
    def test_version_unsupported(self):
        class Stub(ContentManager):
            def __init__(self, ctx):
                self._ctx = ctx

            @requires("1.0.0")
            def fail(self):
                pass

        ctx = MagicMock()
        ctx.version = "0.0.0"
        instance = Stub(ctx)

        with pytest.raises(RuntimeError):
            instance.fail()

    def test_version_supported(self):
        class Stub(ContentManager):
            def __init__(self, ctx):
                self._ctx = ctx

            @requires("1.0.0")
            def success(self):
                pass

        ctx = MagicMock()
        ctx.version = "1.0.0"
        instance = Stub(ctx)

        instance.success()

    def test_version_missing(self):
        class Stub(ContentManager):
            def __init__(self, ctx):
                self._ctx = ctx

            @requires("1.0.0")
            def success(self):
                pass

        ctx = MagicMock()
        ctx.version = None
        instance = Stub(ctx)

        instance.success()


class TestContextVersion:
    @responses.activate
    def test_unknown(self):
        responses.get(
            "http://connect.example/__api__/server_settings",
            json={},
        )

        c = Client("http://connect.example", "12345")
        ctx = c._ctx

        assert ctx.version is None

    @responses.activate
    def test_known(self):
        responses.get(
            "http://connect.example/__api__/server_settings",
            json={"version": "2024.09.24"},
        )

        c = Client("http://connect.example", "12345")
        ctx = c._ctx

        assert ctx.version == "2024.09.24"

    def test_setter(self):
        ctx = Context(Mock())
        ctx.version = "2024.09.24"
        assert ctx.version == "2024.09.24"



from pathlib import Path

import pytest
from packaging import version

from posit import connect

from . import CONNECT_VERSION


class TestContent:
    @classmethod
    def setup_class(cls):
        cls.client = connect.Client()
        cls.content = cls.client.content.create()

    @classmethod
    def teardown_class(cls):
        cls.content.delete()
        assert cls.client.content.count() == 0

    def test_count(self):
        assert self.client.content.count() == 1

    def test_get(self):
        assert self.client.content.get(self.content["guid"]) == self.content

    def test_find(self):
        assert self.client.content.find()

    def test_find_by(self):
        assert self.client.content.find_by(name=self.content["name"]) == self.content

    def test_find_one(self):
        assert self.client.content.find_one()

    def test_content_item_owner(self):
        item = self.client.content.find_one(include=None)
        assert item
        owner = item.owner
        assert owner["guid"] == self.client.me["guid"]

    def test_content_item_owner_from_include(self):
        item = self.client.content.find_one(include="owner")
        assert item
        owner = item.owner
        assert owner["guid"] == self.client.me["guid"]

    @pytest.mark.skipif(
        CONNECT_VERSION <= version.parse("2024.04.1"),
        reason="Python 3.12 not available",
    )
    def test_restart(self):
        # create content
        content = self.client.content.create(name="example-flask-minimal")
        # create bundle
        path = Path("../../../resources/connect/bundles/example-flask-minimal/bundle.tar.gz")
        path = (Path(__file__).parent / path).resolve()
        bundle = content.bundles.create(str(path))
        # deploy bundle
        task = bundle.deploy()
        task.wait_for()
        # restart
        content.restart()
        # delete content
        content.delete()

    @pytest.mark.skipif(
        CONNECT_VERSION <= version.parse("2023.01.1"),
        reason="Quarto not available",
    )
    def test_render(self):
        # create content
        content = self.client.content.create(name="example-quarto-minimal")
        # create bundle
        path = Path("../../../resources/connect/bundles/example-quarto-minimal/bundle.tar.gz")
        path = (Path(__file__).parent / path).resolve()
        bundle = content.bundles.create(str(path))
        # deploy bundle
        task = bundle.deploy()
        task.wait_for()
        # render
        task = content.render()
        task.wait_for()
        # delete content
        content.delete()



"""Environment resources."""

from __future__ import annotations

from abc import abstractmethod
from typing import Protocol

from typing_extensions import (
    List,
    Literal,
    TypedDict,
    runtime_checkable,
)

from .resources import Resource, ResourceSequence

MatchingType = Literal["any", "exact", "none"]
"""Directions for how environments are considered for selection.

    - any: The image may be selected by Connect if not defined in the bundle manifest.
    - exact: The image must be defined in the bundle manifest
    - none: Never use this environment
"""


class Installation(TypedDict):
    """Interpreter installation in an execution environment."""

    path: str
    """The absolute path to the interpreter's executable."""

    version: str
    """The semantic version of the interpreter."""


class Installations(TypedDict):
    """Interpreter installations in an execution environment."""

    installations: List[Installation]
    """Interpreter installations in an execution environment."""


class Environment(Resource):
    @abstractmethod
    def destroy(self) -> None:
        """Destroy the environment.

        Warnings
        --------
        This operation is irreversible.

        Note
        ----
        This action requires administrator privileges.
        """

    @abstractmethod
    def update(
        self,
        *,
        title: str,
        description: str | None = ...,
        matching: MatchingType | None = ...,
        supervisor: str | None = ...,
        python: Installations | None = ...,
        quarto: Installations | None = ...,
        r: Installations | None = ...,
        tensorflow: Installations | None = ...,
    ) -> None:
        """Update the environment.

        Parameters
        ----------
        title : str
            A human-readable title.
        description : str | None, optional, not required
            A human-readable description.
        matching : MatchingType, optional, not required
            Directions for how the environment is considered for selection
        supervisor : str | None, optional, not required
            Path to the supervisor script.
        python : Installations, optional, not required
            The Python installations available in this environment
        quarto : Installations, optional, not required
            The Quarto installations available in this environment
        r : Installations, optional, not required
            The R installations available in this environment
        tensorflow : Installations, optional, not required
            The Tensorflow installations available in this environment

        Note
        ----
        This action requires administrator privileges.
        """


@runtime_checkable
class Environments(ResourceSequence[Environment], Protocol):
    def create(
        self,
        *,
        title: str,
        name: str,
        cluster_name: str | Literal["Kubernetes"],
        matching: MatchingType = "any",
        description: str | None = ...,
        supervisor: str | None = ...,
        python: Installations | None = ...,
        quarto: Installations | None = ...,
        r: Installations | None = ...,
        tensorflow: Installations | None = ...,
    ) -> Environment:
        """Create an environment.

        Parameters
        ----------
        title : str
            A human-readable title.
        name : str
            The container image name used for execution in this environment.
        cluster_name : str | Literal["Kubernetes"]
            The cluster identifier for this environment. Defaults to "Kubernetes" when Off-Host-Execution is enabled.
        description : str, optional
            A human-readable description.
        matching : MatchingType
            Directions for how the environment is considered for selection, by default is "any".
        supervisor : str, optional
            Path to the supervisor script
        python : Installations, optional
            The Python installations available in this environment
        quarto : Installations, optional
            The Quarto installations available in this environment
        r : Installations, optional
            The R installations available in this environment
        tensorflow : Installations, optional
            The Tensorflow installations available in this environment

        Returns
        -------
        Environment

        Note
        ----
        This action requires administrator privileges.
        """
        ...

    def find(self, guid: str, /) -> Environment: ...

    def find_by(
        self,
        *,
        id: str = ...,  # noqa: A002
        guid: str = ...,
        created_time: str = ...,
        updated_time: str = ...,
        title: str = ...,
        name: str = ...,
        description: str | None = ...,
        cluster_name: str | Literal["Kubernetes"] = ...,
        environment_type: str | Literal["Kubernetes"] = ...,
        matching: MatchingType = ...,
        supervisor: str | None = ...,
        python: Installations | None = ...,
        quarto: Installations | None = ...,
        r: Installations | None = ...,
        tensorflow: Installations | None = ...,
    ) -> Environment | None:
        """Find the first record matching the specified conditions.

        There is no implied ordering, so if order matters, you should specify it yourself.

        Parameters
        ----------
        id : str
            The numerical identifier.
        guid : str
            The unique identifier.
        created_time : str
            The timestamp (RFC3339) when the environment was created.
        updated_time : str
            The timestamp (RFC3339) when the environment was updated.
        title : str
            A human-readable title.
        name : str
            The container image name used for execution in this environment.
        description : str, optional
            A human-readable description.
        cluster_name : str | Literal["Kubernetes"]
            The cluster identifier for this environment. Defaults to "Kubernetes" when Off-Host-Execution is enabled.
        environment_type : str | Literal["Kubernetes"]
            The cluster environment type. Defaults to "Kubernetes" when Off-Host-Execution is enabled.
        matching : MatchingType
            Directions for how the environment is considered for selection.
        supervisor : str, optional
            Path to the supervisor script
        python : Installations, optional
            The Python installations available in this environment
        quarto : Installations, optional
            The Quarto installations available in this environment
        r : Installations, optional
            The R installations available in this environment
        tensorflow : Installations, optional
            The Tensorflow installations available in this environment

        Returns
        -------
        Environment | None

        Note
        ----
        This action requires administrator or publisher privileges.
        """



import pytest
import responses
from responses import matchers

# from responses import matchers
from posit.connect.client import Client
from posit.connect.content import ContentItem
from posit.connect.tags import Tag

from .api import load_mock_dict, load_mock_list


class TestTags:
    @responses.activate
    def test_find_all_tags(self):
        """Check GET /v1/tags"""
        # behavior
        mock_get_tags = responses.get(
            "https://connect.example/__api__/v1/tags",
            json=load_mock_list("v1/tags.json"),
        )

        # setup
        client = Client(api_key="12345", url="https://connect.example")

        # invoke
        tags = client.tags.find()

        # assert
        assert mock_get_tags.call_count == 1
        assert len(tags) == 28

        for tag in tags:
            assert isinstance(tag, Tag)

    @responses.activate
    def test_find_tags_by_name(self):
        """Check GET /v1/tags?name=tag_name"""
        # behavior
        mock_get_tags = responses.get(
            "https://connect.example/__api__/v1/tags?name=academy",
            json=load_mock_list("v1/tags?name=academy.json"),
            match=[matchers.query_param_matcher({"name": "academy"})],
        )

        # setup
        client = Client(api_key="12345", url="https://connect.example")

        # invoke
        tags = client.tags.find(name="academy")

        # assert
        assert mock_get_tags.call_count == 1
        assert len(tags) == 1

        for tag in tags:
            assert isinstance(tag, Tag)

    @responses.activate
    def test_find_tags_by_parent(self):
        """Check GET /v1/tags?parent_id=3"""
        # behavior
        mock_get_tags = responses.get(
            "https://connect.example/__api__/v1/tags?parent_id=3",
            json=load_mock_list("v1/tags?parent_id=3.json"),
            match=[matchers.query_param_matcher({"parent_id": "3"})],
        )

        # setup
        client = Client(api_key="12345", url="https://connect.example")

        # invoke
        by_str_tags = client.tags.find(parent_id="3")
        parent_tag = Tag(client._ctx, "/v1/tags/3", id="3", name="Parent")
        by_tag_tags = client.tags.find(parent=parent_tag)

        # assert
        assert mock_get_tags.call_count == 2
        assert len(by_str_tags) == 7
        assert len(by_tag_tags) == 7

    @responses.activate
    def test_get(self):
        # behavior
        mock_get_tag = responses.get(
            "https://connect.example/__api__/v1/tags/33",
            json=load_mock_dict("v1/tags/33.json"),
        )

        # setup
        client = Client(api_key="12345", url="https://connect.example")

        # invoke
        tag = client.tags.get("33")

        # assert
        assert mock_get_tag.call_count == 1
        assert isinstance(tag, Tag)

    @responses.activate
    def test_create_tag(self):
        # behavior
        mock_create_tag = responses.post(
            "https://connect.example/__api__/v1/tags",
            json=load_mock_dict("v1/tags/33.json"),
        )

        # setup
        client = Client(api_key="12345", url="https://connect.example")

        tag = Tag(client._ctx, "/v1/tags/1", id="3", name="Tag")

        # invoke
        academy_tag_parent_id = client.tags.create(name="academy", parent_id=tag["id"])
        academy_tag_parent_tag = client.tags.create(name="academy", parent=tag)

        with pytest.raises(TypeError):
            client.tags.create(
                name="academy",
                parent="not a tag",  # pyright: ignore[reportArgumentType]
            )
        with pytest.raises(ValueError):
            client.tags.create(  # pyright: ignore[reportCallIssue]
                name="academy",
                parent=tag,
                parent_id="asdf",
            )

        # assert
        assert mock_create_tag.call_count == 2

        assert academy_tag_parent_id["id"] == "33"
        assert academy_tag_parent_tag["id"] == "33"


class TestTag:
    @responses.activate
    def test_parent(self):
        # behavior
        mock_get_3_tag = responses.get(
            "https://connect.example/__api__/v1/tags/3",
            json=load_mock_dict("v1/tags/3.json"),
        )
        mock_get_33_tag = responses.get(
            "https://connect.example/__api__/v1/tags/33",
            json=load_mock_dict("v1/tags/33.json"),
        )

        # setup
        client = Client(api_key="12345", url="https://connect.example")

        # invoke
        tag = client.tags.get("33")
        parent_tag = tag.parent_tag

        assert isinstance(parent_tag, Tag)
        parent_parent_tag = parent_tag.parent_tag

        # assert
        assert mock_get_3_tag.call_count == 1
        assert mock_get_33_tag.call_count == 1

        assert parent_parent_tag is None

    @responses.activate
    def test_children(self):
        # behavior
        mock_get_3_tag = responses.get(
            "https://connect.example/__api__/v1/tags/3",
            json=load_mock_dict("v1/tags/3.json"),
        )
        mock_parent_3_tags = responses.get(
            "https://connect.example/__api__/v1/tags",
            json=load_mock_list("v1/tags?parent_id=3.json"),
            match=[matchers.query_param_matcher({"parent_id": "3"})],
        )

        # setup
        client = Client(api_key="12345", url="https://connect.example")

        # invoke
        tag = client.tags.get("3")
        tag_children = tag.child_tags.find()

        # assert
        assert mock_get_3_tag.call_count == 1
        assert mock_parent_3_tags.call_count == 1

        assert len(tag_children) == 7

        for tag_child in tag_children:
            assert isinstance(tag_child, Tag)

    @responses.activate
    def test_descendants(self):
        # behavior
        mock_get_3_tag = responses.get(
            "https://connect.example/__api__/v1/tags/3",
            json=load_mock_dict("v1/tags/3.json"),
        )
        mock_all_tags = responses.get(
            "https://connect.example/__api__/v1/tags",
            json=load_mock_list("v1/tags.json"),
        )

        # setup
        client = Client(api_key="12345", url="https://connect.example")

        # invoke
        tag = client.tags.get("3")
        tag_children = tag.descendant_tags.find()

        # assert
        assert mock_get_3_tag.call_count == 1
        assert mock_all_tags.call_count == 1

        assert len(tag_children) == 21

        for tag_child in tag_children:
            assert isinstance(tag_child, Tag)

    @responses.activate
    def test_content_with_tag(self):
        # behavior
        mock_get_3_tag = responses.get(
            "https://connect.example/__api__/v1/tags/3/content",
            json=load_mock_list("v1/tags/3/content.json"),
        )
        mock_get_33_tag = responses.get(
            "https://connect.example/__api__/v1/tags/33/content",
            json=load_mock_list("v1/tags/33/content.json"),
        )

        # setup
        client = Client(api_key="12345", url="https://connect.example")
        tag3 = Tag(client._ctx, "/v1/tags/3", id="3", name="Tag3")
        tag33 = Tag(client._ctx, "/v1/tags/33", id="33", name="Tag33")

        # invoke
        tag3_content_items = tag3.content_items.find()
        tag33_content_items = tag33.content_items.find()

        # assert
        assert mock_get_3_tag.call_count == 1
        assert mock_get_33_tag.call_count == 1
        assert len(tag3_content_items) == 0
        assert len(tag33_content_items) == 5

        for content_item in tag33_content_items:
            assert isinstance(content_item, ContentItem)

    @responses.activate
    def test_destroy(self):
        # behavior
        mock_29_tag = responses.get(
            "https://connect.example/__api__/v1/tags/29",
            json=load_mock_dict("v1/tags/29.json"),
        )
        mock_29_destroy = responses.delete(
            "https://connect.example/__api__/v1/tags/29",
            json={},  # empty response
        )

        # setup
        client = Client(api_key="12345", url="https://connect.example")

        # invoke
        tag29 = client.tags.get("29")
        tag29.destroy()

        # assert
        assert mock_29_tag.call_count == 1
        assert mock_29_destroy.call_count == 1

    @responses.activate
    def test_update(self):
        # behavior
        mock_get_33_tag = responses.get(
            "https://connect.example/__api__/v1/tags/33",
            json=load_mock_dict("v1/tags/33.json"),
        )
        mock_update_33_tag = responses.patch(
            "https://connect.example/__api__/v1/tags/33",
            json=load_mock_dict("v1/tags/33-patched.json"),
        )

        # setup
        client = Client(api_key="12345", url="https://connect.example")
        tag33 = client.tags.get("33")

        # invoke
        tag33.update(name="academy-updated", parent_id=None)
        tag33.update(name="academy-updated", parent=None)

        parent_tag = Tag(client._ctx, "/v1/tags/1", id="42", name="Parent")
        tag33.update(name="academy-updated", parent=parent_tag)
        tag33.update(name="academy-updated", parent_id=parent_tag["id"])

        # assert
        assert mock_get_33_tag.call_count == 1
        assert mock_update_33_tag.call_count == 4

        # Asserting updated values are deferred to integration testing
        # to avoid agreening with the mocked data


class TestContentItemTags:
    @responses.activate
    def test_find(self):
        # behavior
        content_item_guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        mock_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_item_guid}",
            json=load_mock_dict(f"v1/content/{content_item_guid}.json"),
        )
        mock_tags_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_item_guid}/tags",
            json=load_mock_list(f"v1/content/{content_item_guid}/tags.json"),
        )

        # setup
        client = Client("https://connect.example", "12345")
        content_item = client.content.get(content_item_guid)

        # invoke
        tags = content_item.tags.find()

        # assert
        assert mock_get.call_count == 1
        assert mock_tags_get.call_count == 1
        assert len(tags) == 2

    @responses.activate
    def test_add(self):
        # behavior
        content_item_guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        tag_id = "33"
        mock_content_item_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_item_guid}",
            json=load_mock_dict(f"v1/content/{content_item_guid}.json"),
        )
        mock_tag_get = responses.get(
            f"https://connect.example/__api__/v1/tags/{tag_id}",
            json=load_mock_dict(f"v1/tags/{tag_id}.json"),
        )
        mock_tags_add = responses.post(
            f"https://connect.example/__api__/v1/content/{content_item_guid}/tags",
            json={},  # empty response
        )

        # setup
        client = Client("https://connect.example", "12345")
        content_item = client.content.get(content_item_guid)

        sub_tag = client.tags.get("33")

        # invoke
        content_item.tags.add(sub_tag["id"])
        content_item.tags.add(sub_tag)

        with pytest.raises(TypeError):
            content_item.tags.add(
                123,  # pyright: ignore[reportArgumentType]
            )
        with pytest.raises(ValueError):
            content_item.tags.add("")

        # assert
        assert mock_content_item_get.call_count == 1
        assert mock_tag_get.call_count == 1
        assert mock_tags_add.call_count == 2

    @responses.activate
    def test_delete(self):
        # behavior
        content_item_guid = "f2f37341-e21d-3d80-c698-a935ad614066"
        tag_id = "33"
        mock_content_item_get = responses.get(
            f"https://connect.example/__api__/v1/content/{content_item_guid}",
            json=load_mock_dict(f"v1/content/{content_item_guid}.json"),
        )
        mock_tag_get = responses.get(
            f"https://connect.example/__api__/v1/tags/{tag_id}",
            json=load_mock_dict(f"v1/tags/{tag_id}.json"),
        )
        mock_tags_delete = responses.delete(
            f"https://connect.example/__api__/v1/content/{content_item_guid}/tags/{tag_id}",
            json={},  # empty response
        )

        # setup
        client = Client("https://connect.example", "12345")
        content_item = client.content.get(content_item_guid)

        sub_tag = client.tags.get("33")

        # invoke
        content_item.tags.delete(sub_tag["id"])
        content_item.tags.delete(sub_tag)

        with pytest.raises(TypeError):
            content_item.tags.delete(
                123,  # pyright: ignore[reportArgumentType]
            )
        with pytest.raises(ValueError):
            content_item.tags.delete("")

        # assert
        assert mock_content_item_get.call_count == 1
        assert mock_tag_get.call_count == 1
        assert mock_tags_delete.call_count == 2



from unittest import mock

import pytest
import responses
from responses import matchers

from posit import connect
from posit.connect.metrics import shiny_usage, usage, visits

from ..api import load_mock, load_mock_dict


class TestUsageEventFromEvent:
    def test(self):
        with pytest.raises(TypeError):
            usage.UsageEvent.from_event(
                None  # pyright: ignore[reportArgumentType]
            )


class TestUsageEventFromVisitEvent:
    @classmethod
    def setup_class(cls):
        results = load_mock_dict("v1/instrumentation/content/visits?limit=500.json")["results"]
        assert isinstance(results, list)

        visit_event = visits.VisitEvent(
            mock.Mock(),
            **results[0],
        )
        cls.view_event = usage.UsageEvent.from_visit_event(visit_event)

    def test_content_guid(self):
        assert self.view_event.content_guid == "bd1d2285-6c80-49af-8a83-a200effe3cb3"

    def test_user_guid(self):
        assert self.view_event.user_guid == "08e3a41d-1f8e-47f2-8855-f05ea3b0d4b2"

    def test_variant_key(self):
        assert self.view_event.variant_key == "HidI2Kwq"

    def test_rendering_id(self):
        assert self.view_event.rendering_id == 7

    def test_bundle_id(self):
        assert self.view_event.bundle_id == 33

    def test_started(self):
        assert self.view_event.started == "2018-09-15T18:00:00-05:00"

    def test_ended(self):
        assert self.view_event.ended == "2018-09-15T18:00:00-05:00"

    def test_data_version(self):
        assert self.view_event.data_version == 1

    def test_path(self):
        assert self.view_event.path == "/logs"


class TestUsageEventFromShinyUsageEvent:
    @classmethod
    def setup_class(cls):
        results = load_mock_dict("v1/instrumentation/shiny/usage?limit=500.json")["results"]
        assert isinstance(results, list)
        visit_event = shiny_usage.ShinyUsageEvent(
            mock.Mock(),
            **results[0],
        )
        cls.view_event = usage.UsageEvent.from_shiny_usage_event(visit_event)

    def test_content_guid(self):
        assert self.view_event.content_guid == "bd1d2285-6c80-49af-8a83-a200effe3cb3"

    def test_user_guid(self):
        assert self.view_event.user_guid == "08e3a41d-1f8e-47f2-8855-f05ea3b0d4b2"

    def test_variant_key(self):
        assert self.view_event.variant_key is None

    def test_rendering_id(self):
        assert self.view_event.rendering_id is None

    def test_bundle_id(self):
        assert self.view_event.bundle_id is None

    def test_started(self):
        assert self.view_event.started == "2018-09-15T18:00:00-05:00"

    def test_ended(self):
        assert self.view_event.ended == "2018-09-15T18:01:00-05:00"

    def test_data_version(self):
        assert self.view_event.data_version == 1

    def test_path(self):
        assert self.view_event.path is None


class TestUsageFind:
    @responses.activate
    def test(self):
        # behavior
        mock_get = [
            responses.get(
                "https://connect.example/__api__/v1/instrumentation/content/visits",
                json=load_mock("v1/instrumentation/content/visits?limit=500.json"),
                match=[
                    matchers.query_param_matcher(
                        {
                            "limit": 500,
                        },
                    ),
                ],
            ),
            responses.get(
                "https://connect.example/__api__/v1/instrumentation/content/visits",
                json=load_mock(
                    "v1/instrumentation/content/visits?limit=500&next=23948901087.json"
                ),
                match=[
                    matchers.query_param_matcher(
                        {
                            "next": "23948901087",
                            "limit": 500,
                        },
                    ),
                ],
            ),
            responses.get(
                "https://connect.example/__api__/v1/instrumentation/shiny/usage",
                json=load_mock("v1/instrumentation/shiny/usage?limit=500.json"),
                match=[
                    matchers.query_param_matcher(
                        {
                            "limit": 500,
                        },
                    ),
                ],
            ),
            responses.get(
                "https://connect.example/__api__/v1/instrumentation/shiny/usage",
                json=load_mock("v1/instrumentation/shiny/usage?limit=500&next=23948901087.json"),
                match=[
                    matchers.query_param_matcher(
                        {
                            "next": "23948901087",
                            "limit": 500,
                        },
                    ),
                ],
            ),
        ]

        # setup
        c = connect.Client("https://connect.example", "12345")

        # invoke
        events = c.metrics.usage.find()

        # assert
        assert mock_get[0].call_count == 1
        assert mock_get[1].call_count == 1
        assert mock_get[2].call_count == 1
        assert mock_get[3].call_count == 1
        assert len(events) == 2


class TestUsageFindOne:
    @responses.activate
    def test(self):
        # behavior
        mock_get = [
            responses.get(
                "https://connect.example/__api__/v1/instrumentation/content/visits",
                json=load_mock("v1/instrumentation/content/visits?limit=500.json"),
                match=[
                    matchers.query_param_matcher(
                        {
                            "limit": 500,
                        },
                    ),
                ],
            ),
            responses.get(
                "https://connect.example/__api__/v1/instrumentation/content/visits",
                json=load_mock(
                    "v1/instrumentation/content/visits?limit=500&next=23948901087.json"
                ),
                match=[
                    matchers.query_param_matcher(
                        {
                            "next": "23948901087",
                            "limit": 500,
                        },
                    ),
                ],
            ),
            responses.get(
                "https://connect.example/__api__/v1/instrumentation/shiny/usage",
                json=load_mock("v1/instrumentation/shiny/usage?limit=500.json"),
                match=[
                    matchers.query_param_matcher(
                        {
                            "limit": 500,
                        },
                    ),
                ],
            ),
            responses.get(
                "https://connect.example/__api__/v1/instrumentation/shiny/usage",
                json=load_mock("v1/instrumentation/shiny/usage?limit=500&next=23948901087.json"),
                match=[
                    matchers.query_param_matcher(
                        {
                            "next": "23948901087",
                            "limit": 500,
                        },
                    ),
                ],
            ),
        ]

        # setup
        c = connect.Client("https://connect.example", "12345")

        # invoke
        view_event = c.metrics.usage.find_one()

        # assert
        assert mock_get[0].call_count == 1
        assert mock_get[1].call_count == 0
        assert mock_get[2].call_count == 0
        assert mock_get[3].call_count == 0
        assert view_event
        assert view_event.content_guid == "bd1d2285-6c80-49af-8a83-a200effe3cb3"

    @responses.activate
    def test_none(self):
        # behavior
        mock_get = [
            # return an empty result set to push through the iterator
            responses.get(
                "https://connect.example/__api__/v1/instrumentation/content/visits",
                json=load_mock(
                    "v1/instrumentation/content/visits?limit=500&next=23948901087.json"
                ),
            ),
            responses.get(
                "https://connect.example/__api__/v1/instrumentation/shiny/usage",
                json=load_mock("v1/instrumentation/shiny/usage?limit=500&next=23948901087.json"),
            ),
        ]

        # setup
        c = connect.Client("https://connect.example", "12345")

        # invoke
        view_event = c.metrics.usage.find_one(content_guid="not-found")

        # assert
        assert mock_get[0].call_count == 1
        assert mock_get[1].call_count == 1
        assert view_event is None



import responses

from posit.connect.client import Client

from .api import load_mock


class TestEnvironmentsFind:
    @responses.activate
    def test(self):
        responses.get(
            "https://connect.example/__api__/v1/environments/25438b83-ea6d-4839-ae8e-53c52ac5f9ce",
            json=load_mock(
                "v1/environments/25438b83-ea6d-4839-ae8e-53c52ac5f9ce.json",
            ),
        )
        c = Client("https://connect.example", "12345")
        c._ctx.version = None
        environment = c.environments.find("25438b83-ea6d-4839-ae8e-53c52ac5f9ce")
        assert environment["id"] == "314"


class TestEnvironmentsFindBy:
    @responses.activate
    def test(self):
        responses.get(
            "https://connect.example/__api__/v1/environments",
            json=load_mock(
                "v1/environments.json",
            ),
        )
        c = Client("https://connect.example", "12345")
        c._ctx.version = None
        environment = c.environments.find_by(id="314")
        assert environment
        assert environment["id"] == "314"


class TestEnvironmentsCreate:
    @responses.activate
    def test(self):
        responses.post(
            "https://connect.example/__api__/v1/environments",
            json=load_mock(
                "v1/environments/25438b83-ea6d-4839-ae8e-53c52ac5f9ce.json",
            ),
        )
        c = Client("https://connect.example", "12345")
        c._ctx.version = None
        environment = c.environments.create(
            title="Project Alpha (R 4.1.1, Python 3.10)", name="", cluster_name=""
        )
        assert environment
        assert environment["id"] == "314"


class TestEnvironmentDestroy:
    @responses.activate
    def test(self):
        responses.get(
            "https://connect.example/__api__/v1/environments/25438b83-ea6d-4839-ae8e-53c52ac5f9ce",
            json=load_mock(
                "v1/environments/25438b83-ea6d-4839-ae8e-53c52ac5f9ce.json",
            ),
        )

        mock_delete = responses.delete(
            "https://connect.example/__api__/v1/environments/25438b83-ea6d-4839-ae8e-53c52ac5f9ce",
        )

        c = Client("https://connect.example", "12345")
        c._ctx.version = None
        environment = c.environments.find("25438b83-ea6d-4839-ae8e-53c52ac5f9ce")
        assert environment["id"] == "314"

        environment.destroy()
        assert mock_delete.call_count == 1


class TestEnvironmentUpdate:
    @responses.activate
    def test(self):
        responses.get(
            "https://connect.example/__api__/v1/environments/25438b83-ea6d-4839-ae8e-53c52ac5f9ce",
            json=load_mock(
                "v1/environments/25438b83-ea6d-4839-ae8e-53c52ac5f9ce.json",
            ),
        )

        mock_put = responses.put(
            "https://connect.example/__api__/v1/environments/25438b83-ea6d-4839-ae8e-53c52ac5f9ce",
            json=load_mock(
                "v1/environments/25438b83-ea6d-4839-ae8e-53c52ac5f9ce.json",
            ),
        )

        c = Client("https://connect.example", "12345")
        c._ctx.version = None
        environment = c.environments.find("25438b83-ea6d-4839-ae8e-53c52ac5f9ce")
        assert environment["id"] == "314"

        environment.update(title="test")
        assert mock_put.call_count == 1



from typing import TYPE_CHECKING

import pytest
import responses
from requests.exceptions import HTTPError

from posit.connect.client import Client

from .api import load_mock

if TYPE_CHECKING:
    from posit.connect.jobs import Job, Jobs


class TestJobsMixin:
    @responses.activate
    def test(self):
        responses.get(
            "https://connect.example/__api__/v1/content/f2f37341-e21d-3d80-c698-a935ad614066",
            json=load_mock("v1/content/f2f37341-e21d-3d80-c698-a935ad614066.json"),
        )

        responses.get(
            "https://connect.example/__api__/v1/content/f2f37341-e21d-3d80-c698-a935ad614066/jobs",
            json=load_mock("v1/content/f2f37341-e21d-3d80-c698-a935ad614066/jobs.json"),
        )

        c = Client("https://connect.example", "12345")
        content = c.content.get("f2f37341-e21d-3d80-c698-a935ad614066")
        jobs: Jobs = content.jobs
        assert len(jobs) == 1


class TestJobsFind:
    @responses.activate
    def test(self):
        responses.get(
            "https://connect.example/__api__/v1/content/f2f37341-e21d-3d80-c698-a935ad614066",
            json=load_mock("v1/content/f2f37341-e21d-3d80-c698-a935ad614066.json"),
        )

        responses.get(
            "https://connect.example/__api__/v1/content/f2f37341-e21d-3d80-c698-a935ad614066/jobs/tHawGvHZTosJA2Dx",
            json=load_mock(
                "v1/content/f2f37341-e21d-3d80-c698-a935ad614066/jobs/tHawGvHZTosJA2Dx.json",
            ),
        )

        c = Client("https://connect.example", "12345")
        content = c.content.get("f2f37341-e21d-3d80-c698-a935ad614066")

        job: Job = content.jobs.find("tHawGvHZTosJA2Dx")
        assert job["key"] == "tHawGvHZTosJA2Dx"

    @responses.activate
    def test_miss(self):
        responses.get(
            "https://connect.example/__api__/v1/content/f2f37341-e21d-3d80-c698-a935ad614066",
            json=load_mock("v1/content/f2f37341-e21d-3d80-c698-a935ad614066.json"),
        )

        responses.get(
            "https://connect.example/__api__/v1/content/f2f37341-e21d-3d80-c698-a935ad614066/jobs/not-found",
            status=404,
        )

        c = Client("https://connect.example", "12345")
        content = c.content.get("f2f37341-e21d-3d80-c698-a935ad614066")

        with pytest.raises(HTTPError):
            content.jobs.find("not-found")


class TestJobsFindBy:
    @responses.activate
    def test(self):
        responses.get(
            "https://connect.example/__api__/v1/content/f2f37341-e21d-3d80-c698-a935ad614066",
            json=load_mock("v1/content/f2f37341-e21d-3d80-c698-a935ad614066.json"),
        )

        responses.get(
            "https://connect.example/__api__/v1/content/f2f37341-e21d-3d80-c698-a935ad614066/jobs",
            json=load_mock("v1/content/f2f37341-e21d-3d80-c698-a935ad614066/jobs.json"),
        )

        c = Client("https://connect.example", "12345")
        content = c.content.get("f2f37341-e21d-3d80-c698-a935ad614066")

        job = content.jobs.find_by(key="tHawGvHZTosJA2Dx")
        assert job
        assert job["key"] == "tHawGvHZTosJA2Dx"


class TestJobDestory:
    @responses.activate
    def test(self):
        responses.get(
            "https://connect.example/__api__/v1/content/f2f37341-e21d-3d80-c698-a935ad614066",
            json=load_mock("v1/content/f2f37341-e21d-3d80-c698-a935ad614066.json"),
        )

        responses.get(
            "https://connect.example/__api__/v1/content/f2f37341-e21d-3d80-c698-a935ad614066/jobs/tHawGvHZTosJA2Dx",
            json=load_mock(
                "v1/content/f2f37341-e21d-3d80-c698-a935ad614066/jobs/tHawGvHZTosJA2Dx.json",
            ),
        )

        responses.delete(
            "https://connect.example/__api__/v1/content/f2f37341-e21d-3d80-c698-a935ad614066/jobs/tHawGvHZTosJA2Dx",
        )

        c = Client("https://connect.example", "12345")
        content = c.content.get("f2f37341-e21d-3d80-c698-a935ad614066")

        job = content.jobs.find("tHawGvHZTosJA2Dx")
        job.destroy()



# -*- coding: utf-8 -*-
# mypy: ignore-errors
import os

import pandas as pd
import snowflake.connector
import streamlit as st

from posit.connect.external.snowflake import PositAuthenticator

ACCOUNT = os.getenv("SNOWFLAKE_ACCOUNT")
WAREHOUSE = os.getenv("SNOWFLAKE_WAREHOUSE")

# USER is only required when running the example locally with external browser auth
USER = os.getenv("SNOWFLAKE_USER")

# https://docs.snowflake.com/en/user-guide/sample-data-using
DATABASE = os.getenv("SNOWFLAKE_DATABASE", "snowflake_sample_data")
SCHEMA = os.getenv("SNOWFLAKE_SCHEMA", "tpch_sf1")
TABLE = os.getenv("SNOWFLAKE_TABLE", "lineitem")

session_token = st.context.headers.get("Posit-Connect-User-Session-Token")
auth = PositAuthenticator(local_authenticator="EXTERNALBROWSER", user_session_token=session_token)

con = snowflake.connector.connect(
    user=USER,
    account=ACCOUNT,
    warehouse=WAREHOUSE,
    database=DATABASE,
    schema=SCHEMA,
    authenticator=auth.authenticator,
    token=auth.token,
)

snowflake_user = con.cursor().execute("SELECT CURRENT_USER()").fetchone()
st.write(f"Hello, {snowflake_user[0]}!")

with st.spinner("Loading data from Snowflake..."):
    df = pd.read_sql_query(f"SELECT * FROM {TABLE} LIMIT 10", con)

st.dataframe(df)



from __future__ import annotations

import os
from typing import TYPE_CHECKING, Optional

from typing_extensions import TypedDict

from ..resources import Resources
from .integrations import Integrations
from .sessions import Sessions

if TYPE_CHECKING:
    from ..context import Context

GRANT_TYPE = "urn:ietf:params:oauth:grant-type:token-exchange"
USER_SESSION_TOKEN_TYPE = "urn:posit:connect:user-session-token"
CONTENT_SESSION_TOKEN_TYPE = "urn:posit:connect:content-session-token"


def _get_content_session_token() -> str:
    """Return the content session token.

    Reads the environment variable 'CONNECT_CONTENT_SESSION_TOKEN'.

    Raises
    ------
        ValueError: If CONNECT_CONTENT_SESSION_TOKEN is not set or invalid

    Returns
    -------
        str
    """
    value = os.environ.get("CONNECT_CONTENT_SESSION_TOKEN")
    if not value:
        raise ValueError(
            "Invalid value for 'CONNECT_CONTENT_SESSION_TOKEN': Must be a non-empty string."
        )
    return value


class OAuth(Resources):
    def __init__(self, ctx: Context, api_key: str) -> None:
        super().__init__(ctx)
        self.api_key = api_key
        self._path = "v1/oauth/integrations/credentials"

    @property
    def integrations(self):
        return Integrations(self._ctx)

    @property
    def sessions(self):
        return Sessions(self._ctx)

    def get_credentials(self, user_session_token: Optional[str] = None) -> Credentials:
        """Perform an oauth credential exchange with a user-session-token."""
        # craft a credential exchange request
        data = {}
        data["grant_type"] = GRANT_TYPE
        data["subject_token_type"] = USER_SESSION_TOKEN_TYPE
        if user_session_token:
            data["subject_token"] = user_session_token

        response = self._ctx.client.post(self._path, data=data)
        return Credentials(**response.json())

    def get_content_credentials(self, content_session_token: Optional[str] = None) -> Credentials:
        """Perform an oauth credential exchange with a content-session-token."""
        # craft a credential exchange request
        data = {}
        data["grant_type"] = GRANT_TYPE
        data["subject_token_type"] = CONTENT_SESSION_TOKEN_TYPE
        data["subject_token"] = content_session_token or _get_content_session_token()

        response = self._ctx.client.post(self._path, data=data)
        return Credentials(**response.json())


class Credentials(TypedDict, total=False):
    access_token: str
    issued_token_type: str
    token_type: str



"""Usage resources."""

from __future__ import annotations

from typing import List, overload

from requests.sessions import Session as Session

from .. import resources
from . import shiny_usage, visits


class UsageEvent(resources.BaseResource):
    @staticmethod
    def from_event(
        event: visits.VisitEvent | shiny_usage.ShinyUsageEvent,
    ) -> UsageEvent:
        if isinstance(event, visits.VisitEvent):
            return UsageEvent.from_visit_event(event)

        if isinstance(event, shiny_usage.ShinyUsageEvent):
            return UsageEvent.from_shiny_usage_event(event)

        raise TypeError

    @staticmethod
    def from_visit_event(event: visits.VisitEvent) -> UsageEvent:
        return UsageEvent(
            event._ctx,
            content_guid=event.content_guid,
            user_guid=event.user_guid,
            variant_key=event.variant_key,
            rendering_id=event.rendering_id,
            bundle_id=event.bundle_id,
            started=event.time,
            ended=event.time,
            data_version=event.data_version,
            path=event.path,
        )

    @staticmethod
    def from_shiny_usage_event(
        event: shiny_usage.ShinyUsageEvent,
    ) -> UsageEvent:
        return UsageEvent(
            event._ctx,
            content_guid=event.content_guid,
            user_guid=event.user_guid,
            variant_key=None,
            rendering_id=None,
            bundle_id=None,
            started=event.started,
            ended=event.ended,
            data_version=event.data_version,
            path=None,
        )

    @property
    def content_guid(self) -> str:
        """The associated unique content identifier.

        Returns
        -------
        str
        """
        return self["content_guid"]

    @property
    def user_guid(self) -> str:
        """The associated unique user identifier.

        Returns
        -------
        str
        """
        return self["user_guid"]

    @property
    def variant_key(self) -> str | None:
        """The variant key associated with the visit.

        Returns
        -------
        str | None
            The variant key, or None if the associated content type is static.
        """
        return self["variant_key"]

    @property
    def rendering_id(self) -> int | None:
        """The render id associated with the visit.

        Returns
        -------
        int | None
            The render id, or None if the associated content type is static.
        """
        return self["rendering_id"]

    @property
    def bundle_id(self) -> int | None:
        """The bundle id associated with the visit.

        Returns
        -------
        int
        """
        return self["bundle_id"]

    @property
    def started(self) -> str:
        """The visit timestamp.

        Returns
        -------
        str
        """
        return self["started"]

    @property
    def ended(self) -> str:
        """The visit timestamp.

        Returns
        -------
        str
        """
        return self["ended"]

    @property
    def data_version(self) -> int:
        """The data version.

        Returns
        -------
        int
        """
        return self["data_version"]

    @property
    def path(self) -> str | None:
        """The path requested by the user.

        Returns
        -------
        str
        """
        return self["path"]


class Usage(resources.Resources):
    """Usage resource."""

    @overload
    def find(
        self,
        *,
        content_guid: str = ...,
        min_data_version: int = ...,
        start: str = ...,
        end: str = ...,
    ) -> List[UsageEvent]:
        """Find view events.

        Parameters
        ----------
        content_guid : str, optional
            Filter by an associated unique content identifer, by default ...
        min_data_version : int, optional
            Filter by a minimum data version, by default ...
        start : str, optional
            Filter by the start time, by default ...
        end : str, optional
            Filter by the end time, by default ...

        Returns
        -------
        List[UsageEvent]
        """

    @overload
    def find(self, **kwargs) -> List[UsageEvent]:
        """Find view events.

        Returns
        -------
        List[UsageEvent]
        """

    def find(self, **kwargs) -> List[UsageEvent]:
        """Find view events.

        Returns
        -------
        List[UsageEvent]
        """
        events = []
        finders = (visits.Visits, shiny_usage.ShinyUsage)
        for finder in finders:
            instance = finder(self._ctx)
            events.extend(
                [UsageEvent.from_event(event) for event in instance.find(**kwargs)],
            )
        return events

    @overload
    def find_one(
        self,
        *,
        content_guid: str = ...,
        min_data_version: int = ...,
        start: str = ...,
        end: str = ...,
    ) -> UsageEvent | None:
        """Find a view event.

        Parameters
        ----------
        content_guid : str, optional
            Filter by an associated unique content identifer, by default ...
        min_data_version : int, optional
            Filter by a minimum data version, by default ...
        start : str, optional
            Filter by the start time, by default ...
        end : str, optional
            Filter by the end time, by default ...

        Returns
        -------
        Visit | None
        """

    @overload
    def find_one(self, **kwargs) -> UsageEvent | None:
        """Find a view event.

        Returns
        -------
        Visit | None
        """

    def find_one(self, **kwargs) -> UsageEvent | None:
        """Find a view event.

        Returns
        -------
        UsageEvent | None
        """
        finders = (visits.Visits, shiny_usage.ShinyUsage)
        for finder in finders:
            instance = finder(self._ctx)
            event = instance.find_one(**kwargs)
            if event:
                return UsageEvent.from_event(event)
        return None



from unittest.mock import Mock

import responses
from responses.matchers import json_params_matcher

from posit import connect
from posit.connect.vanities import Vanities, Vanity, VanityMixin


class TestVanityDestroy:
    @responses.activate
    def test_destroy_sends_delete_request(self):
        content_guid = "8ce6eaca-60af-4c2f-93a0-f5f3cddf5ee5"
        base_url = "https://connect.example/__api__"
        endpoint = f"{base_url}/v1/content/{content_guid}/vanity"
        mock_delete = responses.delete(endpoint)

        c = connect.Client("https://connect.example", "12345")
        vanity = Vanity(c._ctx, content_guid=content_guid, path=Mock(), created_time=Mock())

        vanity.destroy()

        assert mock_delete.call_count == 1

    @responses.activate
    def test_destroy_calls_after_destroy_callback(self):
        content_guid = "8ce6eaca-60af-4c2f-93a0-f5f3cddf5ee5"
        base_url = "https://connect.example/__api__"
        endpoint = f"{base_url}/v1/content/{content_guid}/vanity"
        responses.delete(endpoint)

        c = connect.Client("https://connect.example", "12345")
        after_destroy = Mock()
        vanity = Vanity(
            c._ctx,
            after_destroy=after_destroy,
            content_guid=content_guid,
            path=Mock(),
            created_time=Mock(),
        )

        vanity.destroy()

        assert after_destroy.call_count == 1


class TestVanitiesAll:
    @responses.activate
    def test_all_sends_get_request(self):
        base_url = "https://connect.example/__api__"
        endpoint = f"{base_url}/v1/vanities"
        mock_get = responses.get(endpoint, json=[])

        c = connect.Client("https://connect.example", "12345")
        vanities = Vanities(c._ctx)

        vanities.all()

        assert mock_get.call_count == 1


class TestVanityMixin:
    @responses.activate
    def test_vanity_getter_returns_vanity(self):
        guid = "8ce6eaca-60af-4c2f-93a0-f5f3cddf5ee5"
        base_url = "https://connect.example/__api__"
        endpoint = f"{base_url}/v1/content/{guid}/vanity"
        mock_get = responses.get(endpoint, json={"content_guid": guid, "path": "my-dashboard"})

        c = connect.Client("https://connect.example", "12345")
        content = VanityMixin(c._ctx, guid=guid)

        assert content.vanity == "my-dashboard"
        assert mock_get.call_count == 1

    @responses.activate
    def test_vanity_setter_with_string(self):
        guid = "8ce6eaca-60af-4c2f-93a0-f5f3cddf5ee5"
        base_url = "https://connect.example/__api__"
        endpoint = f"{base_url}/v1/content/{guid}/vanity"
        path = "example"
        mock_put = responses.put(
            endpoint,
            json={"content_guid": guid, "path": path},
            match=[json_params_matcher({"path": path})],
        )

        c = connect.Client("https://connect.example", "12345")
        content = VanityMixin(c._ctx, guid=guid)
        content.vanity = path
        assert content.vanity == path

        assert mock_put.call_count == 1

    @responses.activate
    def test_vanity_deleter(self):
        guid = "8ce6eaca-60af-4c2f-93a0-f5f3cddf5ee5"
        base_url = "https://connect.example/__api__"
        endpoint = f"{base_url}/v1/content/{guid}/vanity"
        mock_delete = responses.delete(endpoint)

        c = connect.Client("https://connect.example", "12345")
        content = VanityMixin(c._ctx, guid=guid)
        content._vanity = Vanity(c._ctx, path=Mock(), content_guid=guid, created_time=Mock())
        del content.vanity

        assert content._vanity is None
        assert mock_delete.call_count == 1



"""Permission resources."""

from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional, overload

from requests.sessions import Session as Session

from .resources import BaseResource, Resources

if TYPE_CHECKING:
    from .context import Context
    from .groups import Group
    from .users import User


class Permission(BaseResource):
    def destroy(self) -> None:
        """Destroy the permission."""
        path = f"v1/content/{self['content_guid']}/permissions/{self['id']}"
        self._ctx.client.delete(path)

    @overload
    def update(self, *args, role: str, **kwargs) -> None:
        """Update the permission.

        Parameters
        ----------
        role : str
            The principal role.
        """

    @overload
    def update(self, *args, **kwargs) -> None:
        """Update the permission."""

    def update(self, *args, **kwargs) -> None:
        """Update the permission."""
        body = {
            "principal_guid": self.get("principal_guid"),
            "principal_type": self.get("principal_type"),
            "role": self.get("role"),
        }
        body.update(dict(*args))
        body.update(**kwargs)
        path = f"v1/content/{self['content_guid']}/permissions/{self['id']}"
        response = self._ctx.client.put(path, json=body)
        super().update(**response.json())


class Permissions(Resources):
    def __init__(self, ctx: Context, content_guid: str) -> None:
        super().__init__(ctx)
        self.content_guid = content_guid

    def count(self) -> int:
        """Count the number of permissions.

        Returns
        -------
        int
        """
        return len(self.find())

    @overload
    def create(self, /, *, principal_guid: str, principal_type: str, role: str) -> Permission: ...

    @overload
    def create(self, principal: User | Group, /, *, role: str) -> Permission: ...

    def create(
        self,
        principal: Optional[User | Group] = None,
        /,
        **kwargs,
    ) -> Permission:
        """Create a permission.

        Parameters
        ----------
        principal : User | Group
            The principal user or group to add.
        role : str
            The principal role. Currently only `"viewer"` and `"owner"` are supported.
        principal_guid : str
            User guid or Group guid.
        principal_type : str
            The principal type. Either `"user"` or `"group"`.
        role : str
            The principal role. Currently only `"viewer"` and `"owner"` are supported

        Returns
        -------
        Permission
            The created permission.

        Examples
        --------
        ```python
        from posit import connect

        client = connect.Client()
        content_item = client.content.get(content_guid)

        # New permission role
        role = "viewer"  # Or "owner"

        # Example groups and users
        groups = client.groups.find(prefix="GROUP_NAME_PREFIX_HERE")
        group = groups[0]
        user = client.users.get("USER_GUID_HERE")
        users_and_groups = [user, *groups]

        # Add a group permission
        content_item.permissions.create(group, role=role)
        # Add a user permission
        content_item.permissions.create(user, role=role)

        # Add many group and user permissions with the same role
        for principal in users_and_groups:
            content_item.permissions.create(principal, role=role)

        # Add a group permission manually
        content_item.permissions.create(
            principal_guid=group["guid"],
            principal_type="group",
            role=role,
        )
        # Add a user permission manually
        content_item.permissions.create(
            principal_guid=user["guid"],
            principal_type="user",
            role=role,
        )

        # Confirm new permissions
        content_item.permissions.find()
        ```
        """
        if principal is not None:
            # Avoid circular imports
            from .groups import Group
            from .users import User

            if isinstance(principal, User):
                principal_type = "user"
            elif isinstance(principal, Group):
                principal_type = "group"
            else:
                raise TypeError(f"Invalid argument type: {type(principal).__name__}")

            if "principal_guid" in kwargs:
                raise ValueError("'principal_guid' can not be defined with `principal` present.")
            if "principal_type" in kwargs:
                raise ValueError("'principal_guid' can not be defined with `principal` present.")

            # Set the corresponding kwargs
            kwargs["principal_guid"] = principal["guid"]
            kwargs["principal_type"] = principal_type

        path = f"v1/content/{self.content_guid}/permissions"
        response = self._ctx.client.post(path, json=kwargs)
        return Permission(self._ctx, **response.json())

    def find(self, **kwargs) -> List[Permission]:
        """Find permissions.

        Returns
        -------
        List[Permission]
        """
        path = f"v1/content/{self.content_guid}/permissions"
        response = self._ctx.client.get(path)
        kwargs_items = kwargs.items()
        results = response.json()
        if len(kwargs_items) > 0:
            results = [
                result
                for result in results
                if isinstance(result, dict) and (result.items() >= kwargs_items)
            ]
        return [Permission(self._ctx, **result) for result in results]

    def find_one(self, **kwargs) -> Permission | None:
        """Find a permission.

        Returns
        -------
        Permission | None
        """
        permissions = self.find(**kwargs)
        return next(iter(permissions), None)

    def get(self, uid: str) -> Permission:
        """Get a permission.

        Parameters
        ----------
        uid : str
            The permission id.

        Returns
        -------
        Permission
        """
        path = f"v1/content/{self.content_guid}/permissions/{uid}"
        response = self._ctx.client.get(path)
        return Permission(self._ctx, **response.json())

    def destroy(self, permission: str | Group | User | Permission, /) -> None:
        """Remove supplied content item permission.

        Removes provided permission from the content item's permissions.

        Parameters
        ----------
        permission : str | Group | User | Permission
            The content item permission to remove. If a `str` is received, it is compared against
            the `Permissions`'s `principal_guid`. If a `Group` or `User` is received, the associated
            `Permission` will be removed.

        Examples
        --------
        ```python
        from posit import connect

        #### User-defined inputs ####
        # 1. specify the guid for the content item
        content_guid = "CONTENT_GUID_HERE"
        # 2. specify either the principal_guid or group name prefix
        principal_guid = "USER_OR_GROUP_GUID_HERE"
        group_name_prefix = "GROUP_NAME_PREFIX_HERE"
        ############################

        client = connect.Client()
        content_item = client.content.get(content_guid)

        # Remove a single permission by principal_guid
        content_item.permissions.destroy(principal_guid)

        # Remove by user (if principal_guid is a user)
        user = client.users.get(principal_guid)
        content_item.permissions.destroy(user)

        # Remove by group (if principal_guid is a group)
        group = client.groups.get(principal_guid)
        content_item.permissions.destroy(group)

        # Remove all groups with a matching prefix name
        groups = client.groups.find(prefix=group_name_prefix)
        for group in groups:
            content_item.permissions.destroy(group)

        # Confirm new permissions
        content_item.permissions.find()
        ```
        """
        from .groups import Group
        from .users import User

        if isinstance(permission, str):
            permission_obj = self.get(permission)
        elif isinstance(permission, (Group, User)):
            principal_guid: str = permission["guid"]
            permission_obj = self.find_one(
                principal_guid=principal_guid,
            )
            if permission_obj is None:
                raise ValueError(f"Permission with principal_guid '{principal_guid}' not found.")
        elif isinstance(permission, Permission):
            permission_obj = permission
        else:
            raise TypeError(
                f"destroy() expected `permission=` to have type `str | User | Group | Permission`. Received `{permission}` of type `{type(permission)}`.",
            )

        permission_obj.destroy()



# -*- coding: utf-8 -*-
# mypy: ignore-errors
import os

import pandas as pd
import streamlit as st
from databricks import sql
from databricks.sdk.core import ApiClient, Config, databricks_cli
from databricks.sdk.service.iam import CurrentUserAPI

from posit.connect.external.databricks import PositCredentialsStrategy

DATABRICKS_HOST = os.getenv("DATABRICKS_HOST")
DATABRICKS_HOST_URL = f"https://{DATABRICKS_HOST}"
SQL_HTTP_PATH = os.getenv("DATABRICKS_PATH")

session_token = st.context.headers.get("Posit-Connect-User-Session-Token")
posit_strategy = PositCredentialsStrategy(
    local_strategy=databricks_cli,
    user_session_token=session_token,
)
cfg = Config(
    host=DATABRICKS_HOST_URL,
    # uses Posit's custom credential_strategy if running on Connect,
    # otherwise falls back to the strategy defined by local_strategy
    credentials_strategy=posit_strategy,
)

databricks_user = CurrentUserAPI(ApiClient(cfg)).me()
st.write(f"Hello, {databricks_user.display_name}!")

with sql.connect(
    server_hostname=DATABRICKS_HOST,
    http_path=SQL_HTTP_PATH,
    # https://github.com/databricks/databricks-sql-python/issues/148#issuecomment-2271561365
    credentials_provider=posit_strategy.sql_credentials_provider(cfg),
) as connection:
    with connection.cursor() as cursor:
        cursor.execute("SELECT * FROM samples.nyctaxi.trips LIMIT 10;")
        result = cursor.fetchall()
        st.table(pd.DataFrame(result))



from __future__ import annotations

from typing import Any


def update_dict_values(obj: dict[str, Any], /, **kwargs: Any) -> None:
    """
    Update the values of a dictionary.

    This helper method exists as a workaround for the `dict.update` method. Sometimes, `super()` does not return the `dict` class and. If `super().update(**kwargs)` is called unintended behavior will occur.

    Therefore, this helper method exists to update the `dict`'s values.

    Parameters
    ----------
    obj : dict[str, Any]
        The object to update.
    kwargs : Any
        The key-value pairs to update the object with.

    See Also
    --------
    * https://github.com/posit-dev/posit-sdk-py/pull/366#discussion_r1887845267
    """
    # Could also be performed with:
    # for key, value in kwargs.items():
    #     obj[key] = value

    # Use the `dict` class to explicity update the object in-place
    dict.update(obj, **kwargs)



from posit import connect


class TestVanities:
    @classmethod
    def setup_class(cls):
        cls.client = connect.Client()

    @classmethod
    def teardown_class(cls):
        assert cls.client.content.count() == 0

    def test_all(self):
        content = self.client.content.create(name="example")

        # None by default
        vanities = self.client.vanities.all()
        assert len(vanities) == 0

        # Set
        content.vanity = "example"

        # Get
        vanities = self.client.vanities.all()
        assert len(vanities) == 1

        # Cleanup
        content.delete()

        vanities = self.client.vanities.all()
        assert len(vanities) == 0

    def test_property(self):
        content = self.client.content.create(name="example")

        # None by default
        assert content.vanity is None

        # Set
        content.vanity = "example"

        # Get
        vanity = content.vanity
        assert vanity == "/example/"

        # Delete
        del content.vanity
        assert content.vanity is None

        # Cleanup
        content.delete()

    def test_destroy(self):
        content = self.client.content.create(name="example")

        # None by default
        assert content.vanity is None

        # Set
        content.vanity = "example"

        # Get
        vanity = content.find_vanity()
        assert vanity
        assert vanity["path"] == "/example/"

        # Delete
        vanity.destroy()
        content.reset_vanity()
        assert content.vanity is None

        # Cleanup
        content.delete()



import base64
from typing import Dict
from unittest.mock import patch

import pytest
import responses

from posit.connect import Client
from posit.connect.external.databricks import (
    POSIT_OAUTH_INTEGRATION_AUTH_TYPE,
    CredentialsProvider,
    CredentialsStrategy,
    PositContentCredentialsProvider,
    PositContentCredentialsStrategy,
    PositCredentialsProvider,
    PositCredentialsStrategy,
    PositLocalContentCredentialsProvider,
    PositLocalContentCredentialsStrategy,
    _get_auth_type,
    _new_bearer_authorization_header,
)
from posit.connect.oauth import Credentials


class mock_strategy(CredentialsStrategy):
    def auth_type(self) -> str:
        return "local"

    def __call__(self) -> CredentialsProvider:
        def inner() -> Dict[str, str]:
            return {"Authorization": "Bearer static-pat-token"}

        return inner


def register_mocks():
    responses.post(
        "https://connect.example/__api__/v1/oauth/integrations/credentials",
        match=[
            responses.matchers.urlencoded_params_matcher(
                {
                    "grant_type": "urn:ietf:params:oauth:grant-type:token-exchange",
                    "subject_token_type": "urn:posit:connect:user-session-token",
                    "subject_token": "cit",
                },
            ),
        ],
        json={
            "access_token": "dynamic-viewer-access-token",
            "issued_token_type": "urn:ietf:params:oauth:token-type:access_token",
            "token_type": "Bearer",
        },
    )

    responses.post(
        "https://connect.example/__api__/v1/oauth/integrations/credentials",
        match=[
            responses.matchers.urlencoded_params_matcher(
                {
                    "grant_type": "urn:ietf:params:oauth:grant-type:token-exchange",
                    "subject_token_type": "urn:posit:connect:content-session-token",
                    "subject_token": "cit",
                },
            ),
        ],
        json={
            "access_token": "content-access-token",
            "issued_token_type": "urn:ietf:params:oauth:token-type:access_token",
            "token_type": "Bearer",
        },
    )


class TestPositCredentialsHelpers:
    def test_new_bearer_authorization_header(self):
        credential = Credentials()
        credential["token_type"] = "token_type"
        credential["issued_token_type"] = "issued_token_type"

        with pytest.raises(ValueError):
            _new_bearer_authorization_header(credential)

        credential["access_token"] = "access_token"
        result = _new_bearer_authorization_header(credential)
        assert result == {"Authorization": "Bearer access_token"}

    def test_get_auth_type_local(self):
        assert _get_auth_type("local-auth") == "local-auth"

    @patch.dict("os.environ", {"RSTUDIO_PRODUCT": "CONNECT"})
    def test_get_auth_type_connect(self):
        assert _get_auth_type("local-auth") == POSIT_OAUTH_INTEGRATION_AUTH_TYPE

    @responses.activate
    def test_local_content_credentials_provider(self):
        token_url = "https://my-token/url"
        client_id = "client_id"
        client_secret = "client_secret_123"
        basic_auth = f"{client_id}:{client_secret}"
        b64_basic_auth = base64.b64encode(basic_auth.encode("utf-8")).decode("utf-8")

        responses.post(
            token_url,
            match=[
                responses.matchers.urlencoded_params_matcher(
                    {
                        "grant_type": "client_credentials",
                        "scope": "all-apis",
                    },
                ),
                responses.matchers.header_matcher({"Authorization": f"Basic {b64_basic_auth}"}),
            ],
            json={
                "access_token": "oauth2-m2m-access-token",
                "token_type": "Bearer",
                "expires_in": 3600,
            },
        )

        cp = PositLocalContentCredentialsProvider(token_url, client_id, client_secret)
        assert cp() == {"Authorization": "Bearer oauth2-m2m-access-token"}

    @patch.dict("os.environ", {"CONNECT_CONTENT_SESSION_TOKEN": "cit"})
    @responses.activate
    def test_posit_content_credentials_provider(self):
        register_mocks()

        client = Client(api_key="12345", url="https://connect.example/")
        client._ctx.version = None
        cp = PositContentCredentialsProvider(client=client)
        assert cp() == {"Authorization": "Bearer content-access-token"}

    @responses.activate
    def test_posit_credentials_provider(self):
        register_mocks()

        client = Client(api_key="12345", url="https://connect.example/")
        client._ctx.version = None
        cp = PositCredentialsProvider(client=client, user_session_token="cit")
        assert cp() == {"Authorization": "Bearer dynamic-viewer-access-token"}

    @responses.activate
    def test_local_content_credentials_strategy(self):
        token_url = "https://my-token/url"
        client_id = "client_id"
        client_secret = "client_secret_123"
        basic_auth = f"{client_id}:{client_secret}"
        b64_basic_auth = base64.b64encode(basic_auth.encode("utf-8")).decode("utf-8")

        responses.post(
            token_url,
            match=[
                responses.matchers.urlencoded_params_matcher(
                    {
                        "grant_type": "client_credentials",
                        "scope": "all-apis",
                    },
                ),
                responses.matchers.header_matcher({"Authorization": f"Basic {b64_basic_auth}"}),
            ],
            json={
                "access_token": "oauth2-m2m-access-token",
                "token_type": "Bearer",
                "expires_in": 3600,
            },
        )

        cs = PositLocalContentCredentialsStrategy(
            token_url,
            client_id,
            client_secret,
        )
        cp = cs()
        assert cs.auth_type() == "posit-local-client-credentials"
        assert cp() == {"Authorization": "Bearer oauth2-m2m-access-token"}

    @patch.dict("os.environ", {"CONNECT_CONTENT_SESSION_TOKEN": "cit"})
    @responses.activate
    @patch.dict("os.environ", {"RSTUDIO_PRODUCT": "CONNECT"})
    def test_posit_content_credentials_strategy(self):
        register_mocks()

        client = Client(api_key="12345", url="https://connect.example/")
        client._ctx.version = None
        cs = PositContentCredentialsStrategy(
            local_strategy=mock_strategy(),
            client=client,
        )
        cp = cs()
        assert cs.auth_type() == "posit-oauth-integration"
        assert cp() == {"Authorization": "Bearer content-access-token"}

    @responses.activate
    @patch.dict("os.environ", {"RSTUDIO_PRODUCT": "CONNECT"})
    def test_posit_credentials_strategy(self):
        register_mocks()

        client = Client(api_key="12345", url="https://connect.example/")
        client._ctx.version = None
        cs = PositCredentialsStrategy(
            local_strategy=mock_strategy(),
            user_session_token="cit",
            client=client,
        )
        cp = cs()
        assert cs.auth_type() == "posit-oauth-integration"
        assert cp() == {"Authorization": "Bearer dynamic-viewer-access-token"}

    def test_posit_content_credentials_strategy_fallback(self):
        # local_strategy is used when the content is running locally
        client = Client(api_key="12345", url="https://connect.example/")
        cs = PositContentCredentialsStrategy(
            local_strategy=mock_strategy(),
            client=client,
        )
        cp = cs()
        assert cs.auth_type() == "local"
        assert cp() == {"Authorization": "Bearer static-pat-token"}

    def test_posit_credentials_strategy_fallback(self):
        # local_strategy is used when the content is running locally
        client = Client(api_key="12345", url="https://connect.example/")
        cs = PositCredentialsStrategy(
            local_strategy=mock_strategy(),
            user_session_token="cit",
            client=client,
        )
        cp = cs()
        assert cs.auth_type() == "local"
        assert cp() == {"Authorization": "Bearer static-pat-token"}



from typing import List

from .context import Context
from .resources import BaseResource, Resources
from .tasks import Task


class Variant(BaseResource):
    def render(self) -> Task:
        path = f"variants/{self['id']}/render"
        response = self._ctx.client.post(path)
        return Task(self._ctx, **response.json())


class Variants(Resources):
    def __init__(self, ctx: Context, content_guid: str) -> None:
        super().__init__(ctx)
        self.content_guid = content_guid

    def find(self) -> List[Variant]:
        path = f"applications/{self.content_guid}/variants"
        response = self._ctx.client.get(path)
        results = response.json() or []
        return [Variant(self._ctx, **result) for result in results]



import responses

from posit.connect.client import Client
from posit.connect.system import SystemRuntimeCache
from posit.connect.tasks import Task

from .api import load_mock_dict


class TestSystemCacheRuntime:
    @responses.activate
    def test_runtime_caches(self):
        # behavior
        mock_get_runtimes = responses.get(
            "https://connect.example/__api__/v1/system/caches/runtime",
            json=load_mock_dict("v1/system/caches/runtime.json"),
        )
        mock_delete = responses.delete(
            "https://connect.example/__api__/v1/system/caches/runtime",
            json={"task_id": "12345"},
        )
        mock_task = responses.get(
            "https://connect.example/__api__/v1/tasks/12345",
            json={"task_id": "12345"},
        )

        # setup
        client = Client(api_key="12345", url="https://connect.example")

        # invoke
        runtimes = client.system.caches.runtime.find()

        for runtime in runtimes:
            assert isinstance(runtime, SystemRuntimeCache)
            task = runtime.destroy()
            assert isinstance(task, Task)

        first_runtime = runtimes[0]
        task = first_runtime.destroy()
        assert isinstance(task, Task)
        task = client.system.caches.runtime.destroy(
            language=first_runtime["language"],
            version=first_runtime["version"],
            image_name=first_runtime["image_name"],
        )
        assert isinstance(task, Task)

        # assert
        assert mock_get_runtimes.call_count == 1
        assert mock_delete.call_count == 3
        assert mock_task.call_count == len(runtimes) + 2



from __future__ import annotations

import functools
import weakref
from typing import TYPE_CHECKING, Protocol

from packaging.version import Version

if TYPE_CHECKING:
    from .client import Client


def requires(version: str):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(instance: ContextManager, *args, **kwargs):
            ctx = instance._ctx
            if ctx.version and Version(ctx.version) < Version(version):
                raise RuntimeError(
                    f"This API is not available in Connect version {ctx.version}. Please upgrade to version {version} or later.",
                )
            return func(instance, *args, **kwargs)

        return wrapper

    return decorator


class Context:
    def __init__(self, client: Client):
        # Since this is a child object of the client, we use a weak reference to avoid circular
        # references (which would prevent garbage collection)
        self.client: Client = weakref.proxy(client)

    @property
    def version(self) -> str | None:
        if not hasattr(self, "_version"):
            response = self.client.get("server_settings")
            result = response.json()
            self._version: str | None = result.get("version")

        return self._version

    @version.setter
    def version(self, value: str | None):
        self._version = value


class ContextManager(Protocol):
    _ctx: Context



from unittest.mock import patch

import responses

from posit.connect import Client
from posit.connect.external.snowflake import PositAuthenticator


def register_mocks():
    responses.post(
        "https://connect.example/__api__/v1/oauth/integrations/credentials",
        match=[
            responses.matchers.urlencoded_params_matcher(
                {
                    "grant_type": "urn:ietf:params:oauth:grant-type:token-exchange",
                    "subject_token_type": "urn:posit:connect:user-session-token",
                    "subject_token": "cit",
                },
            ),
        ],
        json={
            "access_token": "dynamic-viewer-access-token",
            "issued_token_type": "urn:ietf:params:oauth:token-type:access_token",
            "token_type": "Bearer",
        },
    )


class TestPositAuthenticator:
    @responses.activate
    @patch.dict("os.environ", {"RSTUDIO_PRODUCT": "CONNECT"})
    def test_posit_authenticator(self):
        register_mocks()

        client = Client(api_key="12345", url="https://connect.example/")
        client._ctx.version = None
        auth = PositAuthenticator(
            local_authenticator="SNOWFLAKE",
            user_session_token="cit",
            client=client,
        )
        assert auth.authenticator == "oauth"
        assert auth.token == "dynamic-viewer-access-token"

    def test_posit_authenticator_fallback(self):
        # local_authenticator is used when the content is running locally
        client = Client(api_key="12345", url="https://connect.example/")
        client._ctx.version = None
        auth = PositAuthenticator(
            local_authenticator="SNOWFLAKE",
            user_session_token="cit",
            client=client,
        )
        assert auth.authenticator == "SNOWFLAKE"
        assert auth.token is None



import pytest
from packaging import version

from posit import connect

from . import CONNECT_VERSION


@pytest.mark.skipif(
    CONNECT_VERSION < version.parse("2023.05.0"),
    reason="Environments API unavailable",
)
class TestEnvironments:
    @classmethod
    def setup_class(cls):
        cls.client = connect.Client()
        cls.environment = cls.client.environments.create(
            title="title",
            name="name",
            cluster_name="Kubernetes",
        )

    @classmethod
    def teardown_class(cls):
        cls.environment.destroy()
        assert len(cls.client.environments) == 0

    def test_find(self):
        uid = self.environment["guid"]
        environment = self.client.environments.find(uid)
        assert environment == self.environment

    def test_find_by(self):
        environment = self.client.environments.find_by(name="name")
        assert environment == self.environment

    def test_update(self):
        assert self.environment["title"] == "title"
        self.environment.update(title="new-title")
        assert self.environment["title"] == "new-title"



"""Package resources."""

from __future__ import annotations

from typing_extensions import (
    Iterable,
    Literal,
    Protocol,
)

from .resources import Resource, ResourceSequence


class ContentPackage(Resource, Protocol):
    pass


class ContentPackages(ResourceSequence[ContentPackage], Protocol):
    def fetch(
        self,
        *,
        language: Literal["python", "r"] = ...,
        name: str = ...,
        version: str = ...,
        hash: str | None = ...,  # noqa: A002
    ) -> Iterable[ContentPackage]:
        """
        Fetch all records matching the specified conditions.

        Parameters
        ----------
        language : {"python", "r"}, not required
            Programming language ecosystem, options are 'python' and 'r'
        name : str, not required
            The package name
        version : str, not required
            The package version
        hash : str or None, optional, not required
            Package description hash for R packages.

        Returns
        -------
        List[ContentPackage]
            The first record matching the specified conditions, or `None` if no such record exists.
        """
        ...

    def find_by(
        self,
        *,
        language: Literal["python", "r"] = ...,
        name: str = ...,
        version: str = ...,
        hash: str | None = ...,  # noqa: A002
    ) -> ContentPackage | None:
        """
        Find the first record matching the specified conditions.

        There is no implied ordering, so if order matters, you should specify it yourself.

        Parameters
        ----------
        language : {"python", "r"}, not required
            Programming language ecosystem, options are 'python' and 'r'
        name : str, not required
            The package name
        version : str, not required
            The package version
        hash : str or None, optional, not required
            Package description hash for R packages.

        Returns
        -------
        ContentPackage | None
            The first record matching the specified conditions, or `None` if no such record exists.
        """
        ...


class Package(Resource, Protocol):
    pass


class Packages(ResourceSequence[Package], Protocol):
    def fetch(
        self,
        *,
        language: Literal["python", "r"] = ...,
        name: str = ...,
        version: str = ...,
        hash: str | None = ...,  # noqa: A002,
        bundle_id: str = ...,
        app_id: str = ...,
        app_guid: str = ...,
    ) -> Iterable[Package]:
        """
        Fetch all records matching the specified conditions.

        Parameters
        ----------
        language : {"python", "r"}, not required
            Programming language ecosystem, options are 'python' and 'r'
        name : str, not required
            The package name
        version : str, not required
            The package version
        hash : str or None, optional, not required
            Package description hash for R packages.
        bundle_id: str, not required
            The unique identifier of the bundle this package is associated with.
        app_id: str, not required
            The numerical identifier of the application this package is associated with.
        app_guid: str, not required
            The unique identifier of the application this package is associated with.

        Returns
        -------
        List[Package]
            The first record matching the specified conditions, or `None` if no such record exists.
        """
        ...

    def find_by(
        self,
        *,
        language: Literal["python", "r"] = ...,
        name: str = ...,
        version: str = ...,
        hash: str | None = ...,  # noqa: A002,
        bundle_id: str = ...,
        app_id: str = ...,
        app_guid: str = ...,
    ) -> Package | None:
        """
        Find the first record matching the specified conditions.

        There is no implied ordering, so if order matters, you should specify it yourself.

        Parameters
        ----------
        language : {"python", "r"}, not required
            Programming language ecosystem, options are 'python' and 'r'
        name : str, not required
            The package name
        version : str, not required
            The package version
        hash : str or None, optional, not required
            Package description hash for R packages.
        bundle_id: str, not required
            The unique identifier of the bundle this package is associated with.
        app_id: str, not required
            The numerical identifier of the application this package is associated with.
        app_guid: str, not required
            The unique identifier of the application this package is associated with.

        Returns
        -------
        Package | None
            The first record matching the specified conditions, or `None` if no such record exists.
        """
        ...
